import os

import pandas as pd
import plotly.express as px

from .figure import Figure


class PandasScatterMap(Figure):
    """
    A special figure used to display a scatter map.
    """

    def __init__(
        self,
        id: str,
        key: str,
        mapbox_token: str,
        dataset_config: dict = None,
        **kwargs
    ):
        dataset_config = dataset_config or None
        super().__init__(id, keys=[key], **kwargs)
        self.dataset_config = dataset_config
        self.mapbox_token = mapbox_token

    def _process_data(self, data):
        return data[self.keys[0]]

    def display(self, df):
        px.set_mapbox_access_token(self.mapbox_token)
        if df.empty:
            dummy = pd.DataFrame(dict(lat=[0], lon=[0]))
            return px.scatter_mapbox(
                dummy, "lat", "lon", mapbox_style="white-bg"
            ).to_dict()
        fig = px.scatter_mapbox(df, **self.dataset_config)
        fig_dict = fig.to_dict()
        return fig_dict



import codecs
import os
import pickle
from collections import deque
from datetime import datetime
from tempfile import TemporaryDirectory, TemporaryFile
from uuid import uuid4
from typing import Iterable

from dash.dependencies import Input, Output
from dash_html_components import Div
from dash_infra.core import Callback, Component, as_callback
from flask import current_app as app
from flask import request


class Storage(Component):
    """
    A storage component modifies a callback so
    it writes into a single serialized dictionary.
    This dictionary (state) has one key per registered
    callback, storing the output of the callback.
    """

    def __init__(self, id: str, use_inner_store=False):
        super().__init__(id)
        self.use_inner_store = use_inner_store
        self.keys = []
        self.clients = []
        self.providers = []

    def layout(self):
        return Div(id=self.id)

    def _serializer(self, client_id: str, callback_id: str, *outputs):
        """
        Serializes the result of a provider callback to the store.
        """
        raise NotImplementedError

    def _deserializer(self, client_id, trigger_keys, states):
        """
        Deserializes the store as an input to a client callback.
        """
        raise NotImplementedError

    def _trigger(self):
        """
        Generates a random string triggering the dependant callbacks.
        """
        return str(uuid4())

    def serializer(self, client_id: str, callback_id: str, *outputs):
        if not outputs:
            outputs = ""

        return self._serializer(client_id, callback_id, *outputs)

    def add_provider(self, callback: Callback, copy=False):
        """
        Registers a callback as a provider to the store.
        It first creates a pre-hook that stores the client ID into the callback state.
        It then creates a post-hook that injects the store's serializer into the callback.
        """
        callback.is_provider[self.id] = True

        def pre_hook(state, client_id, *inputs):
            state.client_id = client_id
            return inputs

        def post_hook(state, callback_output):
            if callback_output is not None:
                return self.serializer(state.client_id, callback.name, callback_output)
            else:
                return {}

        callback._pre_function_hooks.insert(0, pre_hook)
        callback._post_function_hooks.append(post_hook)
        self.providers.append(callback)

        return callback

    def _add_store_prehook(self, callback: Callback):
        """
        Registers a callback as a client to the store.
        It injects a pre-hook that deserializes the content
        of the store as an additional input to the callback.
        """
        if not callback.is_provider[self.id]:

            def pre_hook(s, client_id, *triggers_states_inputs):
                num_triggers = len(callback.dependencies)
                trigger_keys = callback.dependencies

                if self.use_inner_store:
                    states = triggers_states_inputs[num_triggers : 2 * num_triggers]
                    other_inputs = triggers_states_inputs[2 * num_triggers :]
                else:
                    states = []
                    other_inputs = triggers_states_inputs[num_triggers:]

                result = self._deserializer(client_id, trigger_keys, states)

                if callback.unpack_args:
                    return tuple([r for r in result.values()] + list(other_inputs))
                else:
                    return tuple([result] + list(other_inputs))

        else:

            def pre_hook(s, *triggers_states_inputs):
                num_triggers = len(callback.dependencies)
                trigger_keys = callback.dependencies

                if self.use_inner_store:
                    states = triggers_states_inputs[num_triggers : 2 * num_triggers]
                    other_inputs = triggers_states_inputs[2 * num_triggers :]
                else:
                    states = []
                    other_inputs = triggers_states_inputs[num_triggers:]

                result = self._deserializer(callback.client_id, trigger_keys, states)
                if callback.unpack_args:
                    return tuple([r for r in result.values()] + list(other_inputs))
                else:
                    return tuple([result] + list(other_inputs))

        callback._pre_function_hooks.append(pre_hook)

    def add_client(self, callback, unpack_args=False):
        callback.unpack_args = unpack_args
        self.clients.append(callback)

    def _before_registry(self, app):
        """
        This hook is called before the component is registered to the store.
        It configures the inputs and outputs for all client callbacks and provider
        callbacks, so they can use the store.
        """

        super()._before_registry(app)
        for provider in self.providers:
            provider.prepend_input(("main-div", "title"))
            provider.set_outputs((self.id, f"trigger-{provider.name}"))
            self.keys.append(provider.name)

            if self.use_inner_store:
                provider.add_outputs([(self.id, f"value-{provider.name}")])

        for client in self.clients:
            if not client.dependencies:
                client.dependencies = self.keys

            old_inputs = client.dash_inputs
            client.set_inputs(
                [("main-div", "title")]
                + [(self.id, f"trigger-{key}") for key in client.dependencies]
            )
            client.dash_inputs = client.dash_inputs + [
                i for i in old_inputs if i not in client.dash_inputs
            ]

            if self.use_inner_store:
                client.set_states(
                    [
                        (self.id, f"value-{key}")
                        for key in client.dependencies
                        if key != client.name
                    ]
                )

            self._add_store_prehook(client)

    def register_callback(self, callback, **kwargs):
        return self.add_provider(callback, **kwargs)


class PickleStorage(Storage):
    """
    A storage that stores the output of the function as a pickled string
    in a hidden div.
    """
    def __init__(self, id):
        super().__init__(id, use_inner_store=True)

    def _serializer(self, client_id, callback_id, *outputs):
        if len(outputs) == 1:
            outputs = outputs[0]

        serialized = codecs.encode(pickle.dumps(outputs), "base64").decode()
        return self._trigger(), serialized

    def _deserializer(self, client_id, trigger_names, states):
        obj = {}
        for name, state in zip(trigger_names, states):
            obj[name] = pickle.loads(codecs.decode(state.encode(), "base64"))
        return obj


class FileStorage(Storage):
    """
    A storage that stores the output of the function as a pickled
    temporary file.
    """
    import tempfile

    def __init__(self, id, maxsize=20):
        super().__init__(id)
        self.stores = deque(maxlen=maxsize)
        self.maxsize = maxsize
        self._directory = TemporaryDirectory()
        self.directory = self._directory.name

    def get_store(self, client_id):
        if client_id not in self.stores:
            if len(self.stores) == self.maxsize:
                last_id = self.stores.popleft()
                os.remove(os.path.join(self.directory, last_id))
            self.stores.append(client_id)

        store_path = os.path.join(self.directory, client_id)
        return store_path

    def _serializer(self, client_id, callback_id, *outputs):
        path = self.get_store(client_id)
        with open(path, "w+b") as fp:
            if len(outputs) == 1:
                outputs = outputs[0]

            if os.path.getsize(path) > 0:
                state = pickle.load(fp)
                state[callback_id] = outputs
            else:
                state = {callback_id: outputs}

            fp.seek(0)
            pickle.dump(state, fp)

        return self._trigger()

    def _deserializer(self, client_id, trigger_names, _):
        path = self.get_store(client_id)
        with open(path, "rb") as fp:
            obj = pickle.load(fp)
            return {name: obj[name] for name in trigger_names}


class ContextStorage(Storage):
    """
    Stores the output of the function in the application server context.
    """
    def __init__(self, id, maxsize=20):
        super().__init__(id, use_inner_store=False)
        self.stores = deque(maxlen=maxsize)
        self.maxsize = maxsize

    def _before_registry(self, app):
        super()._before_registry(app)
        app.server.store = {}

    def get_store(self, client_id):
        stores = app.store
        if client_id not in self.stores:
            stores[client_id] = {}
            if len(self.stores) == self.maxsize:
                last_id = self.stores.popleft()
                del stores[last_id]
            self.stores.append(client_id)

        return stores[client_id]

    def _serializer(self, client_id, callback_id, *outputs):
        store = self.get_store(client_id)
        if len(outputs) == 1:
            outputs = outputs[0]
        store[callback_id] = outputs
        return self._trigger()

    def _deserializer(self, client_id, trigger_names, _):
        store = self.get_store(client_id)
        return {name: store.get(name) for name in trigger_names}



import base64
import io
import os

import pandas as pd
from dash_bootstrap_components import Alert
from dash_core_components import Upload, Markdown
from dash_html_components import Div
from dash_infra import ComponentGroup, as_callback, as_component
from dash_infra.components.storage import ContextStorage, PickleStorage


class FileInput(ComponentGroup):
    def __init__(self, id, *parsers, multiple=False, store=None, alert=None, **kwargs):
        # fmt: off
        self.id = id
        self.multiple = multiple
        # components
        self.store = store or ContextStorage(f"{id}-store")
        message = "Select a file" if not self.multiple else "Select a list of files"
        self.input = as_component(
            Upload, 
            id=id, 
            children=Div(message, className="col s12"),
            className="file-input row",
            multiple=multiple
        )

        # fmt: on
        self.alert = alert or as_component(
            Alert,
            id=f"alert-{id}",
            color="success",
            dismissable=True,
            duration=2000,
            is_open=False,
        )
        # fmt: off
        #callbacks
        self.parse_to_store = (
            as_callback(
                inputs=(id, "contents"), 
                states=[(id, "filename"), (id, "last_modified")],
            )(self._parse_to_store)
                .to(self.store)
        )

        self.show_alert = as_callback(
            inputs=[(self.store.id, f"trigger-{id}-container")], 
            outputs=[(self.alert.id, "children"), (self.alert.id, "is_open")],
            states=[(id, "filename")]
        )(self._show_alert)

        # fmt: on
        self.parsers = parsers

        if store:
            if alert:
                super().__init__(f"{id}-container", self.input, **kwargs)
            else:
                super().__init__(f"{id}-container", self.alert, self.input, **kwargs)
        else:
            if alert:
                super().__init__(f"{id}-container", self.input, self.store, **kwargs)
            else:
                super().__init__(
                    f"{id}-container", self.alert, self.input, self.store, **kwargs
                )

        self.parse_to_store.func.__name__ = self.id

    def _parse_file_to_store(self, contents, filename):
        _, content_string = contents.split(",")
        decoded = base64.b64decode(content_string)
        extension = os.path.splitext(filename)[1].strip(".")
        for parser in self.parsers:
            func = getattr(parser, extension, None)
            if func:
                return func(decoded)

        return ""

    def _parse_to_store(self, content_list, name_list, _):
        if content_list is not None:
            if not self.multiple:
                return self._parse_file_to_store(content_list, name_list)

            else:
                return [
                    self._parse_file_to_store(c, f)
                    for c, f in zip(content_list, name_list)
                ]

    def _show_alert(self, _, filename):
        if not filename:
            return "", False
        if isinstance(filename, str):
            return f"{filename} uploaded", True
        else:
            filelist = [f"* {fname}" for fname in filename]
            message = "Files uploaded:\n\n" + "\n".join(filelist)
            return Markdown(message), True


class DataFrameParsers(object):
    @staticmethod
    def csv(contents):
        return pd.read_csv(io.StringIO(contents.decode()), sep=";")

    @staticmethod
    def xls(contents):
        return pd.read_excel(io.BytesIO(contents))



import os
from collections import deque
from tempfile import TemporaryDirectory
from uuid import uuid4
from zipfile import ZipFile

from dash_html_components import A, Div
from dash_core_components import Loading
from dash_infra.core import Component, as_callback

from flask import send_from_directory


class DownloadZip(Component):
    def __init__(self, id, maxsize=10):
        self._upload_folder = TemporaryDirectory()
        self.upload_folder = self._upload_folder.name
        self.maxsize = maxsize
        self.tmpfiles = deque(maxlen=maxsize)
        super().__init__(id)

    def layout(self):
        return Div(
            id=f"{self.id}-container",
            children=[
                A(
                    id=self.id,
                    children=f"Download {self.id}",
                    href="",
                    className="btn-flat btn-small disabled",
                ),
                Loading(id=f"{self.id}-loading", fullscreen=True),
            ],
        )

    def _before_registry(self, app):
        def send(filename):
            return send_from_directory(self.upload_folder, filename)

        send.__name__ = f"send-{self.id}"
        app.server.route(f"/download/{self.id}/<filename>")(send)

    def _enable_link(self, href):
        if href:
            return "btn-flat btn-small"
        else:
            return "btn-flat btn-small disabled"

    def register_callback(self, callback, copy=False):
        def post_hook(state, callback_output):
            if len(self.tmpfiles) == self.maxsize:
                last_id = self.tmpfiles.popleft()
                os.remove(os.path.join(self.upload_folder, last_id))

            file_id = str(uuid4()) + ".zip"
            self.tmpfiles.append(file_id)

            with ZipFile(os.path.join(self.upload_folder, file_id), "w") as zipf:
                for fname, fcontent in callback_output.items():
                    with zipf.open(fname, "w") as fp:
                        fp.write(fcontent.encode("utf-8"))

            return f"/download/{self.id}/{file_id}", True

        callback._post_function_hooks.append(post_hook)
        callback.set_outputs([(self.id, "href"), (f"{self.id}-loading", "children")])

        enable_link = as_callback(
            inputs=[(f"{self.id}-loading", "children")], outputs=(self.id, "className"),
        )(self._enable_link)

        self.callbacks.add(callback)
        self.callbacks.add(enable_link)



from typing import Callable

from dash_infra.core import Component
from dash_infra.html_helpers.divs import Row


class ComponentGroup(Component):
    """
    An abstraction on components allowing to group them in
    a common container.
    The container is a Callable returning an HTML component.
    """

    def __init__(
        self, id: str, *components: str, container: Callable = Row
    ) -> "ComponentGroup":
        super().__init__(id)
        self.components = components
        self.container = container

    def layout(self):
        return self.container([c.layout() for c in self.components], id=self.id)
